#include "stdafx.h"
#include "Form1.h"
#include <windows.h>
#include <stdio.h>
#include <atlstr.h>
#include <shellapi.h>

using namespace VFSEditor;

typedef DWORD (CALLBACK* LP_OpenVFS)(const char *, const char *);
typedef DWORD (CALLBACK* LP_CloseVFS)(DWORD);
typedef DWORD (CALLBACK* LP_VGetVfsCount)(DWORD handle);
typedef void  (CALLBACK* LP_VGetVfsNames)(DWORD handle, char **array, DWORD, DWORD maxlen);
typedef DWORD (CALLBACK* LP_VGetFileCount)(DWORD handle, const char *);
typedef void  (CALLBACK* LP_VGetFileNames)(DWORD handle, const char *name, char **array, DWORD, DWORD maxlen);
typedef DWORD (CALLBACK* LP_VOpenFile)(const char *, DWORD handle);
typedef DWORD (CALLBACK* LP_VRemoveFile)( DWORD handle, const char *);
typedef void  (CALLBACK* LP_VCloseFile)(DWORD filehandle);
// typedef DWORD (CALLBACK* LP_VGetFileLength)(DWORD handle, const char *);
typedef DWORD (CALLBACK* LP_VFRead)(void *buffer, DWORD size, DWORD count, DWORD filehandle);
// typedef DWORD (CALLBACK* LP_VFSeek)(DWORD filehandle, DWORD offset, DWORD origin);
typedef DWORD (CALLBACK* LP_VFGetSize)(DWORD filehandle);
// typedef DWORD (CALLBACK* LP_VFGetData)(DWORD filehandle, char *buffer);


HMODULE vfs_dll;
LP_OpenVFS OpenVFS;
LP_CloseVFS CloseVFS;
LP_VGetVfsCount VGetVfsCount;
LP_VGetVfsNames VGetVfsNames;
LP_VGetFileCount VGetFileCount;
LP_VGetFileNames VGetFileNames;
LP_VOpenFile VOpenFile;
LP_VRemoveFile VRemoveFile;
LP_VCloseFile VCloseFile;
LP_VFRead VFRead;
// LP_VFSeek VFSeek;
LP_VFGetSize VFGetSize;
//LP_VFGetData VFGetData;


int APIENTRY _tWinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPTSTR    lpCmdLine,
                     int       nCmdShow)
{
	System::Threading::Thread::CurrentThread->ApartmentState = System::Threading::ApartmentState::STA;
	Application::Run(new Form1());
	return 0;
}


System::Void Form1::OpenVFSDLL()
{
	vfs_dll = LoadLibrary("TriggerVFS");
	
	if( vfs_dll )
	{
		OpenVFS = (LP_OpenVFS)GetProcAddress(vfs_dll, "_OpenVFS@8");
		CloseVFS = (LP_CloseVFS)GetProcAddress(vfs_dll, "_CloseVFS@4");
		VGetVfsCount = (LP_VGetVfsCount)GetProcAddress(vfs_dll, "_VGetVfsCount@4");
		VGetVfsNames = (LP_VGetVfsNames)GetProcAddress(vfs_dll, "_VGetVfsNames@16");
		VGetFileCount = (LP_VGetFileCount)GetProcAddress(vfs_dll, "_VGetFileCount@8");
		VGetFileNames = (LP_VGetFileNames)GetProcAddress(vfs_dll, "_VGetFileNames@20");
		VOpenFile = (LP_VOpenFile)GetProcAddress(vfs_dll, "_VOpenFile@8");
		VRemoveFile = (LP_VRemoveFile)GetProcAddress(vfs_dll, "_VRemoveFile@8");
		VCloseFile = (LP_VCloseFile)GetProcAddress(vfs_dll, "_VCloseFile@4");
		VFRead = (LP_VFRead)GetProcAddress(vfs_dll, "_vfread@16");
//		VFSeek = (LP_VFSeek)GetProcAddress(vfs_dll, "_vfseek@12");
		VFGetSize = (LP_VFGetSize)GetProcAddress(vfs_dll, "_vfgetsize@4");
//		VFGetData = (LP_VFGetData)GetProcAddress(vfs_dll, "_vfgetdata@8");

		if(	(OpenVFS) && 
			(VGetVfsCount) &&
			(VGetVfsNames) &&
			(VGetFileCount) &&
			(VGetFileNames) &&
			(VOpenFile) &&
			(VRemoveFile) &&
			(VCloseFile) &&
			(VFRead) &&
			(VFGetSize) &&
			(CloseVFS)
//			(VFSeek)
//			(VFGetData)
		)
		{
			OutPutPanel->AppendText(S"TriggerVFS.dll loaded.\r\n");
		}
		else
		{
			OutPutPanel->AppendText(S"failed to get functions from TriggerVFS.dll.\r\n");
			FreeLibrary(vfs_dll);
			vfs_dll = NULL;
		}
	}
	else
	{
		OutPutPanel->AppendText(S"cannot load TriggerVFS.dll.\r\n");
		vfs_dll = NULL;
	}
}


System::Void Form1::ShowFiles()
{
	this->set_Enabled( false );

	if( vfs_dll )
	{
		const char *data_idx_name = GetIndexFile();
		DWORD data_idx = OpenVFS(data_idx_name ,"r");
		if( data_idx )
		{
			DWORD vfs_count = VGetVfsCount(data_idx);

			OutPutPanel->AppendText(S"VFS = ");
			OutPutPanel->AppendText(System::Int32(vfs_count).ToString());
			OutPutPanel->AppendText(S"\r\n");

			vfs_count--;
			
			VFSFileTree->Nodes->Clear();
			
			if( vfs_count )
			{
				DWORD i,j;
				int k;

				char **vfs_names = new char*[vfs_count];

				for( i = 0; i<vfs_count; i++)
					vfs_names[i] = new char[256];

				VGetVfsNames(data_idx, vfs_names, vfs_count, 255);

				System::Collections::Specialized::HybridDictionary *dict = new System::Collections::Specialized::HybridDictionary();
				System::Windows::Forms::TreeNode *tree = new System::Windows::Forms::TreeNode(S"Root");
				
				tree->Expand();

				for( i=0; i<vfs_count; i++ )
				{
					DWORD file_count = VGetFileCount(data_idx, vfs_names[i]);
					OutPutPanel->AppendText(vfs_names[i]);
					OutPutPanel->AppendText(S" = ");
					OutPutPanel->AppendText(System::Int32(file_count).ToString());
					OutPutPanel->AppendText(S"\r\n");

					char **vfsfile_names = new char*[file_count];

					for( j = 0; j<file_count; j++)
						vfsfile_names[j] = new char[256];

					VGetFileNames(data_idx, vfs_names[i], vfsfile_names, file_count, 255);
					
					for( j=0; j<file_count; j++)
					{	

						String *pathname = new System::String(vfsfile_names[j]);
						String *path [] = pathname->Split( S"\\"->ToCharArray() );

						System::Collections::Specialized::HybridDictionary *r = dict;
						System::Windows::Forms::TreeNode *parent = tree;

						String *key = new String(S"");

						if( path->Count > 0)
						{
							for( k=0; k < path->Count; k++)
							{
								key = String::Concat(key, S"\\", path[k]);
								System::Windows::Forms::TreeNode *node = __try_cast<System::Windows::Forms::TreeNode *>( r->Item[key] );

								if( !node )
								{
									node = new System::Windows::Forms::TreeNode(path[k]);
									parent->Nodes->Add( node );
									dict->Add(key, node );
								}

								parent = node;								
							}
						}
					}
				}

				VFSFileTree->Nodes->Add(tree);
			}
			
			CloseVFS( data_idx );
			OutPutPanel->AppendText(S"ready.\r\n");
		}
		else
		{
			OutPutPanel->AppendText(S"OpenVFS(\"");
			OutPutPanel->AppendText(GetIndexFile());
			OutPutPanel->AppendText("\",r) failed\r\n");
		}
	}
	else
	{
		OutPutPanel->AppendText(S"no DLL loaded.");
	}


	this->set_Enabled( true );
}


public __gc class ExtractAllThread 
{
private:
	static System::Windows::Forms::Form *gui;
	static System::Windows::Forms::TextBox *OutPutPanel;
	static const char *IndexFile;
	static System::String *path;

 public:
	 static void init(System::Windows::Forms::Form *f, System::Windows::Forms::TextBox *outp, const char *idx, System::String *p)
	 {
		 gui = f;
		 OutPutPanel = outp;
		 IndexFile = idx;
		 path = p;
	 }

     static void ThreadProc()
     {

		//gui->set_Enabled( false );

    if( vfs_dll )
	{
		DWORD data_idx = OpenVFS(IndexFile,"r");
		if( data_idx )
		{
			DWORD vfs_count = VGetVfsCount(data_idx);

			OutPutPanel->AppendText(S"VFS = ");
			OutPutPanel->AppendText(System::Int32(vfs_count).ToString());
			OutPutPanel->AppendText(S"\r\n");

			vfs_count--;
			
			if( vfs_count )
			{
				DWORD i,j;
				char **vfs_names = new char*[vfs_count];
				for( i = 0; i<vfs_count; i++)
					vfs_names[i] = new char[256];

				VGetVfsNames(data_idx, vfs_names, vfs_count, 255);
				
				// reopen 
				// CloseVFS( data_idx );
				// data_idx = OpenVFS(GetIndexFile(),"r");

				// vfs_count = 1;
				// vfs_names[0] = "BASIC.VFS";
				for( i=0; i<vfs_count; i++ )
				{
					DWORD file_count = VGetFileCount(data_idx, vfs_names[i]);
					OutPutPanel->AppendText(vfs_names[i]);
					OutPutPanel->AppendText(S" = ");
					OutPutPanel->AppendText(System::Int32(file_count).ToString());
					OutPutPanel->AppendText(S"\r\n");
					
					char **vfsfile_names = new char*[file_count];
					for( j = 0; j<file_count; j++)
						vfsfile_names[j] = new char[256];

					VGetFileNames(data_idx, vfs_names[i], vfsfile_names, file_count, 255);
					
					for( j=0; j<file_count; j++)
					{
						String *filename = String::Concat(path, S"\\", new String(vfsfile_names[j]));

						// OutPutPanel->AppendText(S"save:\"");
						// OutPutPanel->AppendText(filename);
						// OutPutPanel->AppendText(S"\" : ");

						DWORD filehandle = VOpenFile(vfsfile_names[j], data_idx);

						if( filehandle )
						{
							// create dirpatch
							CString cfilename(filename);
							String *relname = new System::String(vfsfile_names[j]);
							String *dirname [] = relname->Split( S"\\"->ToCharArray() );
							relname = path;

							if( dirname->Count > 1 )
							{

								for( int i=0; i<dirname->Count - 1; i++)
								{
									relname = String::Concat(relname, S"\\", dirname[i]);
									CString crelname(relname);
									// OutPutPanel->AppendText(relname);
									// OutPutPanel->AppendText(S"\r\n");
									CreateDirectory(crelname, NULL);
								}
							}

							/* read data */
							DWORD size = VFGetSize(filehandle);

							// OutPutPanel->AppendText(System::Int32(size).ToString());
							// OutPutPanel->AppendText(S"\r\n");
							
							char *buffer = new char[size];
							// VFSeek( filehandle, 0, 0);
							DWORD rread = VFRead( buffer, 1, size, filehandle );
							
							if(rread == size )
							{
								FILE *out = fopen(cfilename, "wb");	
								if( out )
								{
									if( size != fwrite(buffer, 1, size, out))
									{
										OutPutPanel->AppendText(S"write to file failed.\r\n");
									}
									fclose(out);
								}
								else
								{
									OutPutPanel->AppendText(S"cannot open target\r\n");
								}

							}
							else
							{

								OutPutPanel->AppendText(System::Int32(rread).ToString());
								OutPutPanel->AppendText(S"!=");
								OutPutPanel->AppendText(System::Int32(size).ToString());
								OutPutPanel->AppendText(S"\r\n");
								OutPutPanel->AppendText(S"cannot read data\r\n");
							}

							delete [] buffer;
							VCloseFile(filehandle);
						}
						else
						{
							OutPutPanel->AppendText(S"cannot open file\r\n");
							break;
						}
					}

					//OutPutPanel->Clear();
					OutPutPanel->AppendText(S"done.\r\n");
				}
			}
			CloseVFS( data_idx );
			OutPutPanel->AppendText(S"ready.\r\n");
		}
		else
		{
			OutPutPanel->AppendText(S"OpenVFS(\"");
			OutPutPanel->AppendText(IndexFile);
			OutPutPanel->AppendText("\",r) failed\r\n");
		}
	}
	else
	{
		OutPutPanel->AppendText(S"no DLL loaded.");
	}
	
	//gui->set_Enabled( true );

	 }
};


System::Void Form1::ExtractAll(String *path)
{
	ExtractAllThread::init( this, OutPutPanel, GetIndexFile(), path  );
	Thread *t = new Thread(new ThreadStart(0, &ExtractAllThread::ThreadProc));
	t->Start();
	t->Join();
}


System::Boolean Form1::Extract(String *ifilename, String *ofilename)
{
	CString cifilename(ifilename),
			cofilename(ofilename);

	System::Boolean ret = false;

	//this->set_Enabled( false );

	OutPutPanel->AppendText(S"saving \"");
	OutPutPanel->AppendText(ifilename);
	OutPutPanel->AppendText(S"\" to ");
	OutPutPanel->AppendText(ofilename);
	OutPutPanel->AppendText(S"...\r\n");

	if( vfs_dll )
	{
		DWORD data_idx = OpenVFS(GetIndexFile(),"r");

		if( data_idx )
		{
			DWORD filehandle = VOpenFile(cifilename, data_idx);
			if( filehandle )
			{
				/* read data */
				DWORD size = VFGetSize(filehandle);

				char *buffer = new char[size];
				DWORD rread = VFRead( buffer, 1, size, filehandle );
							
				if( rread == size )
				{
					String *dirname [] = ofilename->Split( S"\\"->ToCharArray() );
					String *relname = S"";

					if( dirname->Count > 1 )
					{
						for( int i=0; i<dirname->Count - 1; i++)
						{
							relname = String::Concat(relname, dirname[i], S"\\");
							CString crelname(relname);
							CreateDirectory(crelname, NULL);
						}
					}

					FILE *out = fopen(cofilename, "wb");	
					if( out )
					{
						if( size != fwrite(buffer, 1, size, out))
						{
							OutPutPanel->AppendText(S"write to file failed.\r\n");
						}

						ret = true;
						fclose(out);
					}
					else
					{
						OutPutPanel->AppendText(S"cannot open target\r\n");
					}

				}
				else
				{
					OutPutPanel->AppendText(System::Int32(rread).ToString());
					OutPutPanel->AppendText(S"!=");
					OutPutPanel->AppendText(System::Int32(size).ToString());
					OutPutPanel->AppendText(S"\r\n");
					OutPutPanel->AppendText(S"cannot read data\r\n");
				}

				delete [] buffer;
				VCloseFile(filehandle);
				OutPutPanel->AppendText(S"done.\r\n");
			}
			else
			{
				OutPutPanel->AppendText(S"cannot open file\r\n");
			}

			CloseVFS( data_idx );
			OutPutPanel->AppendText(S"ready.\r\n");
		}
		else
		{
			OutPutPanel->AppendText(S"OpenVFS(\"");
			OutPutPanel->AppendText(GetIndexFile());
			OutPutPanel->AppendText("\",r) failed\r\n");
		}
	}
	else
	{
		OutPutPanel->AppendText(S"no DLL loaded.");
	}

	//this->set_Enabled( true );

	return ret;
}

System::Void Form1::DelFile( String * file )
{
	this->set_Enabled( false );

	OutPutPanel->AppendText(S"deleting \"");
	OutPutPanel->AppendText(file);
	OutPutPanel->AppendText(S"...\r\n");

	if( vfs_dll )
	{
		DWORD data_idx = OpenVFS(GetIndexFile(),"w+");

		if( data_idx )
		{
			VRemoveFile( data_idx, CString(file) );
		}
		else
		{
			OutPutPanel->AppendText(S"OpenVFS(\"");
			OutPutPanel->AppendText(GetIndexFile());
			OutPutPanel->AppendText("\",w+) failed\r\n");
		}
		
		CloseVFS( data_idx );
		OutPutPanel->AppendText(S"ready.\r\n");
	}
	else
	{
		OutPutPanel->AppendText(S"no DLL loaded.");
	}

	this->set_Enabled( true );
}

const char *Form1::GetIndexFile()
{
	CString *a = new CString(dataidxDlg->getString());
	return *a;
}


System::Void Form1::VFSFileTree_DoubleClick(System::Object *  sender, System::EventArgs *  e)
{		
	TreeNode * node = VFSFileTree->SelectedNode;
	if( node )
	{
		String *path = node->FullPath;
		try
		{
			path = path->Substring(5, path->Length - 5);

			#undef GetTempPath
			String *temp_file = String::Concat(System::IO::Path::GetTempPath(), "vfseditor.tmp\\");
			temp_file = String::Concat(temp_file, path);
			CString ctemp_file(temp_file);

			if(Extract(path, temp_file))
			{
				ShellExecute( 0, "open", ctemp_file, NULL, NULL, SW_SHOWNORMAL );
			}
		}
		catch(...)
		{
			OutPutPanel->AppendText(S"failed\r\n");
		}
	}
}

System::Void Form1::RemoveTempDir()
{
	
		//String *temp_dir = String::Concat(System::IO::Path::GetTempPath(), "vfseditor.tmp");
		//System::IO::Directory::Delete(temp_dir);
}